/*
 * Copyright (C) 2013-2015 by the University of Southern California
 * $Id$
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#include "global_var.h"
#include "parse.hh"
#include <getopt.h>
#include "utility.hh"
#include "version.h"
#include <unordered_map>
using namespace std;

#define PASS4           256
#define PASS6           257

bool report_messages = false;
enum report_options report_qry = NO_REPORT;
enum report_options report_rsp = NO_REPORT;
bool report_extra = false;
uint16_t anon_flag = ANON_NONE;
bool verbose_log = false;
string command_line;

/*
	validate parse flag
*/
void valid_parse_flag(string& s)
{
	// We accept "u" to mean unified queries as a special case
	if (s.find_first_not_of("mqruQR") != string::npos) {
		cerr<<"[ERROR] -p only supports m)essages, q)ueries, r)esponses; capitalize QR to include packet information" << endl;
		exit(1);
	}
	if (s.find("u") !=  string::npos) {
		// xxx: backwards compabitility
		report_qry = UNIFIED_REPORT;
	}
	if (s.find("m") !=  string::npos)
		report_messages = true;
	if (s.find("m") !=  string::npos)
		report_extra = false;
	if (s.find("q") !=  string::npos)
		report_qry = NON_UNIFIED_REPORT;
	if (s.find("Q") !=  string::npos)
		report_qry = UNIFIED_REPORT;
	if (s.find("r") !=  string::npos)
		report_rsp = NON_UNIFIED_REPORT;
	if (s.find("R") !=  string::npos)
		report_rsp = UNIFIED_REPORT;
}

/*
     check what to anonymize
*/
uint16_t check_anon_flag(string flag_strs)
{
	uint32_t flag = 0;
	vector<string> flag_vec;
	string all_flag;
	unordered_map<string, unsigned int> flag_map = {
		{"all",         ANON_ALL},         //anon all
		{"addr",        ANON_ADDR},        //server and client addr
		{"server-addr", ANON_SERVER_ADDR}, //only server addr
		{"client-addr", ANON_CLIENT_ADDR}, //only client addr
		{"name",        ANON_NAME},        //all the names(qname, name in rdata, etc.)
		{"id",          ANON_DNS_ID}       //dns id in header
	};
	for (auto it : flag_map) all_flag += "\"" + it.first + "\" ";
	str_split(flag_strs, flag_vec, ',', true);
	for (string s : flag_vec) {
		if (flag_map.find(s) == flag_map.end())
			errx(1, "[error] -a only supports %s", all_flag.c_str());
		flag |= flag_map[s];
	}
	return flag;
}

/*
	output usage
*/
void usage(const char *comm){
	cerr<<" Usage:\n "<<
	  comm << " [-p mqrQR] [-a OPTION] [-k KEYFILE] [--pass4 BITS] [--pass6 BITS]\n"
	  "           [-i PATH] [-o PATH] [-f PREFIX] [-c CMP] [-v]\n"
	  " -p/--parse mqrQR             report DNS m)essages, q)ueries, r)esponses,\n"
	  "                              capital Q and R includes messages in those reports\n"
	  "                              default is mqr (all three, non-unified)\n"
	  " -a/--anon WHAT               anonymize: id,addr,server-addr,client-addr,name,all (separated by comma)\n"
	  "                              id:   DNS id in packet header\n"
	  "                              name: query and response content\n"
	  "                              addr: server address (server-addr) and client address (client-addr)\n"
	  "                                    address with port 53 is identified as server address\n"
	  "                                    if both src and dst port are 53, both of the address are anonymized\n"
	  "                                    anonymized addresses are suffixed with *\n"
	  "                              all:  all of the above\n"
	  " -k/--key KEYFILE             specify the key file. KEYFILE is created if it doesn't exist\n"
	  " --pass4 BITS                 preserve BITS high bits of IPv4 addrsss as un-anonymized;\n"
	  "                              default is 0, anonymizing all bits if anon is set\n"
	  " --pass6 BITS                 similar to --pass4, but for IPv6 address \n"
	  " -i/--input PATH              specify input file path; required\n"
	  " -o/--output-path PATH        specify output path which should be directory; default is current directory\n"
	  "                              use '-' for stdout"
	  " -c/--compressor COMP         compressor, default is '/usr/bin/xz', use \"none\" to disable\n"
	  " -f/--output-prefix PREFIX    prefix of the output file, default is timestamp of the first packet\n"
	  " -v/--verbose                 verbose log; default is none\n"
	  "\n(dnsanon version " << VERSION << ")\n";
	exit(1);
}

/*
	main function
*/
int main(int argc, char *argv[])
{
	const char *comm = argv[0];	//command
	int opt = -1;
	//int func = -1;
	int pass_bits4 = 0, pass_bits6 = 0;
	
	string testfunc="", parse_flag = "", output_prefix = "", compressor = "/usr/bin/xz";
	string inputPath, outputPath, keyFile, tmp;
        
      	//set_anon = false;
      	command_line = "# ";
      	for (int i=0; i<argc; i++) {
      		string tmp = argv[i];
      		command_line += tmp;
      		if (i + 1 != argc)
      			command_line += " ";
      	}
      	command_line += "\n";

	struct option long_options[] = {
		{"anon",          1, NULL, 'a'},
		{"compressor",    1, NULL, 'c'},
		{"extra",         0, NULL, 'e'},
		{"Extra",         0, NULL, 'E'},
		{"output-prefix", 1, NULL, 'f'},
		{"help",          0, NULL, 'h'},
		{"input",         1, NULL, 'i'},
		{"key",           1, NULL, 'k'},
		{"output-path",   1, NULL, 'o'},
		{"parse",         1, NULL, 'p'},
		{"verbose",       0, NULL, 'v'},
		{"pass4",         1, NULL, PASS4},
		{"pass6",         1, NULL, PASS6}
	};

	while ((opt = getopt_long(argc, argv, "a:c:eEf:i:k:o:p:", long_options, NULL)) != EOF) {
		switch(opt) {
		case PASS4:
			pass_bits4 = atoi(optarg);
			if (pass_bits4 < 0 || pass_bits4 > 32)
				errx(1, "[error] --pass4 argument must be within [0, 32]");
			break;
		case PASS6:
			pass_bits6 = atoi(optarg);
			if (pass_bits6 < 0 || pass_bits6 > 128)
				errx(1, "[error] --pass6 argument must be within [0, 128]");
			break;
		case 'a':
			tmp = optarg;
			anon_flag = check_anon_flag(tmp);
			break;
		case 'c':
			compressor = optarg;
			break;

		case 'e':
			// xxx: backwards compatibility, depricated
			report_extra = true;
			break;
		case 'E':
			// xxx: backwards compatibility, depricated
			report_extra = false;
			break;
		case 'f':
			output_prefix = optarg;
			break;
		case 'i':
			inputPath = optarg;
			break;
		case 'k':
			keyFile = optarg;
			break;
		case 'o':
			outputPath = optarg;
			break;
		case 'p':
			parse_flag = optarg;
			valid_parse_flag(parse_flag);
			break;
		case 'v':
			verbose_log = true;
			break;
		default:
			usage(comm);
		}
	}
        argc--, argv++; 		//skip first
        if (argc <= 0) {		//not valid input
            usage(comm);
        }
        
        if (parse_flag.length() == 0) {			//you must provide input
		if (verbose_log) warnx("[warn] parse option is not given, use default mqr");
		parse_flag = "mqr";
		valid_parse_flag(parse_flag);
	}
        if (inputPath.length() == 0) 			//you must provide input
		errx(1, "[error] -i PATH is required");
                
        if (outputPath.length() == 0) {
		if (verbose_log) warnx("[warn] no output path is given, use current directory");
		outputPath = "./";
	}
	
        if (anon_flag != ANON_NONE && (keyFile.length() == 0))	//if -anon, it must have -k
		errx(1, "[error] -k KEYFILE is required for anonymization");
        
        if (anon_flag == ANON_NONE && (keyFile.length() != 0))	//if -k without -anon, -k is ignored
		if (verbose_log) warnx("[warn] no anon flag provided, key is ignored");
        
        //call parse function
	vector<string> tmp_vec;
	str_split(compressor, tmp_vec);
	if (tmp_vec.size() < 1)
		errx(1, "[error] compressor is invalid");

	int32_t pbits = (pass_bits6 << 16) + pass_bits4; //ipv6 uses high 16 bits and ipv4 uses low 16 bits
	parse(inputPath, outputPath, keyFile, pbits, output_prefix, tmp_vec);
        
	return 0;
}

